import functools

from ..engine.opcodes import OpCode
from .                import representation as jar
from ..exceptions     import JertlCodeEmitterException

@functools.singledispatch
def emit(structure, _):
    raise JertlCodeEmitterException(f'emit cannot handle {structure}')

@emit.register(int)
@emit.register(float)
@emit.register(str)
@emit.register(bool)
@emit.register(type(None))
def _(structure, _):
    yield OpCode.MATCH_VALUE, structure

@emit.register(jar.Variable)
def _(variable, bindings):
    identifier = variable.identifier
    if identifier in bindings:
        yield OpCode.MATCH_VARIABLE, identifier
    else:
        yield OpCode.BIND_VARIABLE, identifier
        bindings |= {identifier}

@emit.register(jar.VarArgs)
def _(varargs, bindings):
    identifier = varargs.variable.identifier
    if identifier in bindings:
        yield OpCode.MATCH_VARARGS, identifier
    else:
        yield OpCode.BIND_VARARGS, identifier
        bindings |= {identifier}

@emit.register(dict)
def _(dict_, bindings):
    yield OpCode.MASK_IF_DICT,

    kwargs = None
    for key, structure in dict_.items():
        if type(key) == jar.KWArgs:
            # The KWArgs variable must be processed after the other keys
            kwargs = structure
        else:
            yield OpCode.FOCUS_ON_KEY, key
            yield from emit(structure, bindings)
            yield OpCode.POP_FOCUS,

    if kwargs is not None:
        yield from emit(kwargs, bindings)

@emit.register(list)
def _(array, bindings):
    yield OpCode.MASK_IF_LIST,

    for element in array:
        if isinstance(element, jar.VarArgs):
            yield from emit(element, bindings)
        else:
            yield OpCode.FOCUS_ON_HEAD,
            yield from emit(element, bindings)
            yield OpCode.POP_FOCUS,

    yield OpCode.MATCH_VALUE, []

def optimize_splats(instructions):
    splat = None

    #
    # This is ugly and will remain so intentionally as incentive to use jertl for peephole optimization
    # (AKA the "Eat my own dogfood initiative")
    #
    for instruction in instructions:
        if splat is not None:
            # Previous opcode was a splat.
            # If we are matching focus to the empty list
            # we can optimize away the push to context stack
            if instruction == (OpCode.MATCH_VALUE, []):
                yield OpCode.BIND_VARIABLE, splat[1]
                splat = None
            else:
                yield splat
                splat = None
                if instruction[0] == OpCode.BIND_VARARGS:
                    #
                    # Need to handle the case [[BIND_VARARGS, var1], [BIND_VARARGS, var2], [MATCH_VALUE, []]]
                    #
                    splat = instruction
                else:
                    yield instruction

        elif instruction[0] == OpCode.BIND_VARARGS:
            splat = instruction
        else:
            yield instruction

def emit_match(structure):
    bindings = set()
    yield from optimize_splats(emit(structure, bindings))
    yield OpCode.YIELD_BINDINGS,

def emit_collation(collation, initial_bindings):
    for matcher in collation:
        yield OpCode.FOCUS_ON_BINDING, matcher.variable.identifier
        yield from optimize_splats(emit(matcher.structure, initial_bindings))
        yield OpCode.POP_FOCUS,

    yield OpCode.YIELD_BINDINGS,



import sys
import pathlib
sys.path.append(str(pathlib.Path('..').resolve()))

# Use Semver maintained by hatch
import jertl.__about__

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project   = 'jertl'
copyright = '2022, Ray Pelletier'
author    = 'Ray Pelletier'
release   = jertl.__about__.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ['sphinx.ext.napoleon']

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']
autodoc_member_order = 'bysource'

# Napoleon settings
napoleon_include_init_with_doc = True
napoleon_type_aliases = None

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ['_static']




class JertlSyntaxError(RuntimeError):
    """
    JertlSyntaxError Error in mini-language source
    """


class JertlInterpreterException (Exception):
    pass

class JertlCodeEmitterException (Exception):
    pass

class JertlFillException (Exception):
    """
    JertlFillException Reference to unbound variable during a fill operation
    """

class JertlPurityException (Exception):
    pass



import functools

from ..ast        import representation as jar
from ..exceptions import JertlFillException, JertlPurityException

@functools.singledispatch
def construct(template, _):
    raise JertlFillException(f'build could not handle {template}')

@construct.register(bool)
@construct.register(int)
@construct.register(type(None))
@construct.register(float)
@construct.register(str)
def _(template, _):
    return template

@construct.register(jar.Variable)
def _(template, bindings):
    identifier = template.identifier
    if identifier in bindings:
        return bindings[identifier]
    else:
        raise JertlFillException(f"'{identifier}' not bound")

@construct.register(dict)
def _(template, bindings):
    result = {}
    for key, template in template.items():
        if type(key) == jar.KWArgs:
            result.update(bindings[template.identifier])
        else:
            result[key] = construct(template, bindings)
    #
    return result

@construct.register(list)
def _(template, bindings):
    result = []
    for element in template:
        if isinstance(element, jar.VarArgs):
            result.extend(bindings[element.variable.identifier])
        else:
            result.append(construct(element, bindings))
    return result


@functools.singledispatch
def ensure_purity(value):
    """
    ensure_purity Make sure the data structure we were given is mask free

    Args:
        value: the data structure to inspect

    Raises:
        JertlPurityException: A masked list/dict was found
    """
    raise JertlPurityException(f'purity_check failed {value}')

@ensure_purity.register(bool)
@ensure_purity.register(int)
@ensure_purity.register(type(None))
@ensure_purity.register(float)
@ensure_purity.register(str)
def _(_):
    pass

@ensure_purity.register(list)
def _(value):
    for item in value:
        ensure_purity(item)

@ensure_purity.register(dict)
def _(value):
    for k, v in value.items():
        ensure_purity(k)
        ensure_purity(v)



# Generated from Jertl.g4 by ANTLR 4.11.1
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,20,128,2,0,7,0,2,1,7,1,2,2,7,2,2,3,7,3,2,4,7,4,2,5,7,5,2,6,7,
        6,2,7,7,7,2,8,7,8,2,9,7,9,2,10,7,10,2,11,7,11,2,12,7,12,2,13,7,13,
        2,14,7,14,1,0,1,0,1,0,1,1,1,1,1,1,1,1,1,1,1,2,5,2,40,8,2,10,2,12,
        2,43,9,2,1,2,1,2,1,3,5,3,48,8,3,10,3,12,3,51,9,3,1,3,1,3,5,3,55,
        8,3,10,3,12,3,58,9,3,1,3,1,3,1,4,1,4,1,4,1,4,1,5,1,5,1,5,1,5,1,6,
        1,6,1,6,1,6,3,6,74,8,6,1,7,1,7,1,7,1,7,5,7,80,8,7,10,7,12,7,83,9,
        7,1,7,1,7,3,7,87,8,7,1,7,1,7,1,7,1,7,3,7,93,8,7,1,8,1,8,1,8,1,8,
        1,9,1,9,1,9,1,10,1,10,1,10,1,10,5,10,106,8,10,10,10,12,10,109,9,
        10,1,10,1,10,1,10,1,10,3,10,115,8,10,1,11,1,11,3,11,119,8,11,1,12,
        1,12,1,12,1,13,1,13,1,14,1,14,1,14,0,0,15,0,2,4,6,8,10,12,14,16,
        18,20,22,24,26,28,0,1,2,0,10,13,17,18,124,0,30,1,0,0,0,2,33,1,0,
        0,0,4,41,1,0,0,0,6,49,1,0,0,0,8,61,1,0,0,0,10,65,1,0,0,0,12,73,1,
        0,0,0,14,92,1,0,0,0,16,94,1,0,0,0,18,98,1,0,0,0,20,114,1,0,0,0,22,
        118,1,0,0,0,24,120,1,0,0,0,26,123,1,0,0,0,28,125,1,0,0,0,30,31,3,
        12,6,0,31,32,5,0,0,1,32,1,1,0,0,0,33,34,3,12,6,0,34,35,5,9,0,0,35,
        36,3,12,6,0,36,37,5,0,0,1,37,3,1,0,0,0,38,40,3,8,4,0,39,38,1,0,0,
        0,40,43,1,0,0,0,41,39,1,0,0,0,41,42,1,0,0,0,42,44,1,0,0,0,43,41,
        1,0,0,0,44,45,5,0,0,1,45,5,1,0,0,0,46,48,3,8,4,0,47,46,1,0,0,0,48,
        51,1,0,0,0,49,47,1,0,0,0,49,50,1,0,0,0,50,52,1,0,0,0,51,49,1,0,0,
        0,52,56,5,9,0,0,53,55,3,10,5,0,54,53,1,0,0,0,55,58,1,0,0,0,56,54,
        1,0,0,0,56,57,1,0,0,0,57,59,1,0,0,0,58,56,1,0,0,0,59,60,5,0,0,1,
        60,7,1,0,0,0,61,62,3,28,14,0,62,63,5,14,0,0,63,64,3,12,6,0,64,9,
        1,0,0,0,65,66,3,28,14,0,66,67,5,15,0,0,67,68,3,12,6,0,68,11,1,0,
        0,0,69,74,3,14,7,0,70,74,3,20,10,0,71,74,3,26,13,0,72,74,3,28,14,
        0,73,69,1,0,0,0,73,70,1,0,0,0,73,71,1,0,0,0,73,72,1,0,0,0,74,13,
        1,0,0,0,75,76,5,1,0,0,76,81,3,16,8,0,77,78,5,2,0,0,78,80,3,16,8,
        0,79,77,1,0,0,0,80,83,1,0,0,0,81,79,1,0,0,0,81,82,1,0,0,0,82,86,
        1,0,0,0,83,81,1,0,0,0,84,85,5,2,0,0,85,87,3,18,9,0,86,84,1,0,0,0,
        86,87,1,0,0,0,87,88,1,0,0,0,88,89,5,3,0,0,89,93,1,0,0,0,90,91,5,
        1,0,0,91,93,5,3,0,0,92,75,1,0,0,0,92,90,1,0,0,0,93,15,1,0,0,0,94,
        95,5,13,0,0,95,96,5,4,0,0,96,97,3,12,6,0,97,17,1,0,0,0,98,99,5,5,
        0,0,99,100,3,28,14,0,100,19,1,0,0,0,101,102,5,6,0,0,102,107,3,22,
        11,0,103,104,5,2,0,0,104,106,3,22,11,0,105,103,1,0,0,0,106,109,1,
        0,0,0,107,105,1,0,0,0,107,108,1,0,0,0,108,110,1,0,0,0,109,107,1,
        0,0,0,110,111,5,7,0,0,111,115,1,0,0,0,112,113,5,6,0,0,113,115,5,
        7,0,0,114,101,1,0,0,0,114,112,1,0,0,0,115,21,1,0,0,0,116,119,3,12,
        6,0,117,119,3,24,12,0,118,116,1,0,0,0,118,117,1,0,0,0,119,23,1,0,
        0,0,120,121,5,8,0,0,121,122,3,28,14,0,122,25,1,0,0,0,123,124,7,0,
        0,0,124,27,1,0,0,0,125,126,5,16,0,0,126,29,1,0,0,0,10,41,49,56,73,
        81,86,92,107,114,118
    ]

class JertlParser ( Parser ):

    grammarFileName = "Jertl.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'{'", "','", "'}'", "':'", "'**'", "'['", 
                     "']'", "'*'", "'-->'", "'null'", "'true'", "'false'", 
                     "<INVALID>", "'~'", "':='" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "IMPLIES", "NULL", "TRUE", "FALSE", "STRING", 
                      "MATCHES", "ASSIGNED", "IDENTIFIER", "INTEGER", "FLOAT", 
                      "WS", "LINE_COMMENT" ]

    RULE_toplevel_structure = 0
    RULE_transform = 1
    RULE_collation = 2
    RULE_rule_ = 3
    RULE_matcher = 4
    RULE_setter = 5
    RULE_structure = 6
    RULE_obj = 7
    RULE_key_value = 8
    RULE_kwargs = 9
    RULE_array = 10
    RULE_element = 11
    RULE_varargs = 12
    RULE_atom = 13
    RULE_variable = 14

    ruleNames =  [ "toplevel_structure", "transform", "collation", "rule_", 
                   "matcher", "setter", "structure", "obj", "key_value", 
                   "kwargs", "array", "element", "varargs", "atom", "variable" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    T__4=5
    T__5=6
    T__6=7
    T__7=8
    IMPLIES=9
    NULL=10
    TRUE=11
    FALSE=12
    STRING=13
    MATCHES=14
    ASSIGNED=15
    IDENTIFIER=16
    INTEGER=17
    FLOAT=18
    WS=19
    LINE_COMMENT=20

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.11.1")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class Toplevel_structureContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def structure(self):
            return self.getTypedRuleContext(JertlParser.StructureContext,0)


        def EOF(self):
            return self.getToken(JertlParser.EOF, 0)

        def getRuleIndex(self):
            return JertlParser.RULE_toplevel_structure

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterToplevel_structure" ):
                listener.enterToplevel_structure(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitToplevel_structure" ):
                listener.exitToplevel_structure(self)




    def toplevel_structure(self):

        localctx = JertlParser.Toplevel_structureContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_toplevel_structure)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 30
            self.structure()
            self.state = 31
            self.match(JertlParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class TransformContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def structure(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(JertlParser.StructureContext)
            else:
                return self.getTypedRuleContext(JertlParser.StructureContext,i)


        def IMPLIES(self):
            return self.getToken(JertlParser.IMPLIES, 0)

        def EOF(self):
            return self.getToken(JertlParser.EOF, 0)

        def getRuleIndex(self):
            return JertlParser.RULE_transform

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterTransform" ):
                listener.enterTransform(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitTransform" ):
                listener.exitTransform(self)




    def transform(self):

        localctx = JertlParser.TransformContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_transform)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 33
            self.structure()
            self.state = 34
            self.match(JertlParser.IMPLIES)
            self.state = 35
            self.structure()
            self.state = 36
            self.match(JertlParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class CollationContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EOF(self):
            return self.getToken(JertlParser.EOF, 0)

        def matcher(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(JertlParser.MatcherContext)
            else:
                return self.getTypedRuleContext(JertlParser.MatcherContext,i)


        def getRuleIndex(self):
            return JertlParser.RULE_collation

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCollation" ):
                listener.enterCollation(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCollation" ):
                listener.exitCollation(self)




    def collation(self):

        localctx = JertlParser.CollationContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_collation)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 41
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==16:
                self.state = 38
                self.matcher()
                self.state = 43
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 44
            self.match(JertlParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Rule_Context(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IMPLIES(self):
            return self.getToken(JertlParser.IMPLIES, 0)

        def EOF(self):
            return self.getToken(JertlParser.EOF, 0)

        def matcher(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(JertlParser.MatcherContext)
            else:
                return self.getTypedRuleContext(JertlParser.MatcherContext,i)


        def setter(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(JertlParser.SetterContext)
            else:
                return self.getTypedRuleContext(JertlParser.SetterContext,i)


        def getRuleIndex(self):
            return JertlParser.RULE_rule_

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterRule_" ):
                listener.enterRule_(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitRule_" ):
                listener.exitRule_(self)




    def rule_(self):

        localctx = JertlParser.Rule_Context(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_rule_)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 49
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==16:
                self.state = 46
                self.matcher()
                self.state = 51
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 52
            self.match(JertlParser.IMPLIES)
            self.state = 56
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==16:
                self.state = 53
                self.setter()
                self.state = 58
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 59
            self.match(JertlParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class MatcherContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def variable(self):
            return self.getTypedRuleContext(JertlParser.VariableContext,0)


        def MATCHES(self):
            return self.getToken(JertlParser.MATCHES, 0)

        def structure(self):
            return self.getTypedRuleContext(JertlParser.StructureContext,0)


        def getRuleIndex(self):
            return JertlParser.RULE_matcher

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterMatcher" ):
                listener.enterMatcher(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitMatcher" ):
                listener.exitMatcher(self)




    def matcher(self):

        localctx = JertlParser.MatcherContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_matcher)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 61
            self.variable()
            self.state = 62
            self.match(JertlParser.MATCHES)
            self.state = 63
            self.structure()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class SetterContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def variable(self):
            return self.getTypedRuleContext(JertlParser.VariableContext,0)


        def ASSIGNED(self):
            return self.getToken(JertlParser.ASSIGNED, 0)

        def structure(self):
            return self.getTypedRuleContext(JertlParser.StructureContext,0)


        def getRuleIndex(self):
            return JertlParser.RULE_setter

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSetter" ):
                listener.enterSetter(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSetter" ):
                listener.exitSetter(self)




    def setter(self):

        localctx = JertlParser.SetterContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_setter)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 65
            self.variable()
            self.state = 66
            self.match(JertlParser.ASSIGNED)
            self.state = 67
            self.structure()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class StructureContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def obj(self):
            return self.getTypedRuleContext(JertlParser.ObjContext,0)


        def array(self):
            return self.getTypedRuleContext(JertlParser.ArrayContext,0)


        def atom(self):
            return self.getTypedRuleContext(JertlParser.AtomContext,0)


        def variable(self):
            return self.getTypedRuleContext(JertlParser.VariableContext,0)


        def getRuleIndex(self):
            return JertlParser.RULE_structure

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterStructure" ):
                listener.enterStructure(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitStructure" ):
                listener.exitStructure(self)




    def structure(self):

        localctx = JertlParser.StructureContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_structure)
        try:
            self.state = 73
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [1]:
                self.enterOuterAlt(localctx, 1)
                self.state = 69
                self.obj()
                pass
            elif token in [6]:
                self.enterOuterAlt(localctx, 2)
                self.state = 70
                self.array()
                pass
            elif token in [10, 11, 12, 13, 17, 18]:
                self.enterOuterAlt(localctx, 3)
                self.state = 71
                self.atom()
                pass
            elif token in [16]:
                self.enterOuterAlt(localctx, 4)
                self.state = 72
                self.variable()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ObjContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self._key_value = None # Key_valueContext
            self.key_values = list() # of Key_valueContexts

        def key_value(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(JertlParser.Key_valueContext)
            else:
                return self.getTypedRuleContext(JertlParser.Key_valueContext,i)


        def kwargs(self):
            return self.getTypedRuleContext(JertlParser.KwargsContext,0)


        def getRuleIndex(self):
            return JertlParser.RULE_obj

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterObj" ):
                listener.enterObj(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitObj" ):
                listener.exitObj(self)




    def obj(self):

        localctx = JertlParser.ObjContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_obj)
        self._la = 0 # Token type
        try:
            self.state = 92
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,6,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 75
                self.match(JertlParser.T__0)
                self.state = 76
                localctx._key_value = self.key_value()
                localctx.key_values.append(localctx._key_value)
                self.state = 81
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,4,self._ctx)
                while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                    if _alt==1:
                        self.state = 77
                        self.match(JertlParser.T__1)
                        self.state = 78
                        localctx._key_value = self.key_value()
                        localctx.key_values.append(localctx._key_value) 
                    self.state = 83
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input,4,self._ctx)

                self.state = 86
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==2:
                    self.state = 84
                    self.match(JertlParser.T__1)
                    self.state = 85
                    self.kwargs()


                self.state = 88
                self.match(JertlParser.T__2)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 90
                self.match(JertlParser.T__0)
                self.state = 91
                self.match(JertlParser.T__2)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Key_valueContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STRING(self):
            return self.getToken(JertlParser.STRING, 0)

        def structure(self):
            return self.getTypedRuleContext(JertlParser.StructureContext,0)


        def getRuleIndex(self):
            return JertlParser.RULE_key_value

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterKey_value" ):
                listener.enterKey_value(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitKey_value" ):
                listener.exitKey_value(self)




    def key_value(self):

        localctx = JertlParser.Key_valueContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_key_value)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 94
            self.match(JertlParser.STRING)
            self.state = 95
            self.match(JertlParser.T__3)
            self.state = 96
            self.structure()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class KwargsContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def variable(self):
            return self.getTypedRuleContext(JertlParser.VariableContext,0)


        def getRuleIndex(self):
            return JertlParser.RULE_kwargs

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterKwargs" ):
                listener.enterKwargs(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitKwargs" ):
                listener.exitKwargs(self)




    def kwargs(self):

        localctx = JertlParser.KwargsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_kwargs)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 98
            self.match(JertlParser.T__4)
            self.state = 99
            self.variable()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ArrayContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self._element = None # ElementContext
            self.elements = list() # of ElementContexts

        def element(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(JertlParser.ElementContext)
            else:
                return self.getTypedRuleContext(JertlParser.ElementContext,i)


        def getRuleIndex(self):
            return JertlParser.RULE_array

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterArray" ):
                listener.enterArray(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitArray" ):
                listener.exitArray(self)




    def array(self):

        localctx = JertlParser.ArrayContext(self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_array)
        self._la = 0 # Token type
        try:
            self.state = 114
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,8,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 101
                self.match(JertlParser.T__5)
                self.state = 102
                localctx._element = self.element()
                localctx.elements.append(localctx._element)
                self.state = 107
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==2:
                    self.state = 103
                    self.match(JertlParser.T__1)
                    self.state = 104
                    localctx._element = self.element()
                    localctx.elements.append(localctx._element)
                    self.state = 109
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 110
                self.match(JertlParser.T__6)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 112
                self.match(JertlParser.T__5)
                self.state = 113
                self.match(JertlParser.T__6)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ElementContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def structure(self):
            return self.getTypedRuleContext(JertlParser.StructureContext,0)


        def varargs(self):
            return self.getTypedRuleContext(JertlParser.VarargsContext,0)


        def getRuleIndex(self):
            return JertlParser.RULE_element

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterElement" ):
                listener.enterElement(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitElement" ):
                listener.exitElement(self)




    def element(self):

        localctx = JertlParser.ElementContext(self, self._ctx, self.state)
        self.enterRule(localctx, 22, self.RULE_element)
        try:
            self.state = 118
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [1, 6, 10, 11, 12, 13, 16, 17, 18]:
                self.enterOuterAlt(localctx, 1)
                self.state = 116
                self.structure()
                pass
            elif token in [8]:
                self.enterOuterAlt(localctx, 2)
                self.state = 117
                self.varargs()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class VarargsContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def variable(self):
            return self.getTypedRuleContext(JertlParser.VariableContext,0)


        def getRuleIndex(self):
            return JertlParser.RULE_varargs

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterVarargs" ):
                listener.enterVarargs(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitVarargs" ):
                listener.exitVarargs(self)




    def varargs(self):

        localctx = JertlParser.VarargsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 24, self.RULE_varargs)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 120
            self.match(JertlParser.T__7)
            self.state = 121
            self.variable()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class AtomContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NULL(self):
            return self.getToken(JertlParser.NULL, 0)

        def TRUE(self):
            return self.getToken(JertlParser.TRUE, 0)

        def FALSE(self):
            return self.getToken(JertlParser.FALSE, 0)

        def INTEGER(self):
            return self.getToken(JertlParser.INTEGER, 0)

        def FLOAT(self):
            return self.getToken(JertlParser.FLOAT, 0)

        def STRING(self):
            return self.getToken(JertlParser.STRING, 0)

        def getRuleIndex(self):
            return JertlParser.RULE_atom

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAtom" ):
                listener.enterAtom(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAtom" ):
                listener.exitAtom(self)




    def atom(self):

        localctx = JertlParser.AtomContext(self, self._ctx, self.state)
        self.enterRule(localctx, 26, self.RULE_atom)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 123
            _la = self._input.LA(1)
            if not(((_la) & ~0x3f) == 0 and ((1 << _la) & 408576) != 0):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class VariableContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IDENTIFIER(self):
            return self.getToken(JertlParser.IDENTIFIER, 0)

        def getRuleIndex(self):
            return JertlParser.RULE_variable

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterVariable" ):
                listener.enterVariable(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitVariable" ):
                listener.exitVariable(self)




    def variable(self):

        localctx = JertlParser.VariableContext(self, self._ctx, self.state)
        self.enterRule(localctx, 28, self.RULE_variable)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 125
            self.match(JertlParser.IDENTIFIER)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








from   dataclasses import dataclass
from   functools   import singledispatch
from   typing      import Dict, List, Any
from   copy        import copy

from   jertl.engine.opcodes   import OpCode
from   jertl.engine.masks     import MaskedList, MaskedDict, freeze, snapshot
from   jertl.exceptions       import JertlInterpreterException

@dataclass
class Context:
    identifier:  str
    bindings:    Dict[str, Any]
    focus_stack: List[Any]
    ic:          int

@singledispatch
def values_match(a, b):
    return a == b

@values_match.register(int)
def _(a, b):
    return isinstance(b, int) and not isinstance(b, bool) and a == b

@values_match.register(float)
def _(a, b):
    return isinstance(b, float) and a == b

@values_match.register(bool)
def _(a, b):
    return isinstance(b, bool) and a == b

@values_match.register(type(None))
def _(_, b):
    return b is None


class Interpreter:
    """_summary_"""

    def __init__(self, instructions, trace=False):
        """__init__

        Args:
            instructions List[OpCode]: the 'program' to interpred
            trace (bool, optional): Flag determining if we should trace operation of vm, defaults to False
        """
        self.instructions  = instructions
        self.trace         = trace
        self.focus_stack   = []
        self.context_stack = []
        self.ic            = 0
        self.running       = False

    def backtrack(self):
        if self.trace:
            print(f'...backtracking')

        if len(self.context_stack) > 0:
            context = self.context_stack[-1]
            if len(context.focus_stack[-1]) == 0:
                self.context_stack.pop()
                self.backtrack()
            else:
                context.bindings[context.identifier].widen()
                context.focus_stack[-1].pop()
                self.focus_stack = copy(context.focus_stack)
                self.focus_stack[-1] = context.focus_stack[-1].snapshot()
                self.ic = context.ic
                self.bindings = copy(self.bindings)
        else:
            # Nothing left to do
            self.running = False

    def maybe_trace(self):
        if self.trace:
            ic    = self.ic
            fc    = len(self.focus_stack) - 1
            focus = self.focus_stack[-1] if fc >= 0 else '<NO-FOCUS>'
            print(f'ic: {ic:3}, focus[{fc}]:{self.instructions[ic]} -- {focus}')


    def match_all(self, focus, bindings=None):
        self.running = True
        self.bindings = {} if bindings is None else bindings
        self.focus_stack.append(focus)

        while self.running:
            self.maybe_trace()

            instruction = self.instructions[self.ic]
            opcode      = instruction[0]
            self.ic    += 1

            if opcode == OpCode.MATCH_VALUE:
                if not values_match(instruction[1], self.focus_stack[-1]):
                    self.backtrack()

            elif opcode == OpCode.MATCH_VARIABLE:
                if not values_match(self.bindings[instruction[1]], self.focus_stack[-1]):
                    self.backtrack()

            elif opcode == OpCode.BIND_VARIABLE:
                self.bindings[instruction[1]] = self.focus_stack[-1]

            elif opcode == OpCode.BIND_VARARGS:
                binding = self.focus_stack[-1].snapshot()
                binding.collapse()
                self.bindings[instruction[1]] = binding

                context = Context(identifier=instruction[1],
                                  focus_stack=[snapshot(focus) for focus in self.focus_stack],
                                  bindings=copy(self.bindings),
                                  ic=self.ic)

                self.context_stack.append(context)

            elif opcode == OpCode.MATCH_VARARGS:
                binding = self.bindings[instruction[1]]
                if len(binding) <= len(self.focus_stack[-1]):
                    for value in binding:
                        if not values_match(value, self.focus_stack[-1].pop()):
                            self.backtrack()
                            break
                else:
                    self.backtrack()

            elif opcode == OpCode.MASK_IF_LIST:
                if not isinstance(self.focus_stack[-1], list):
                    self.backtrack()
                else:
                    self.focus_stack[-1] = MaskedList(self.focus_stack[-1])

            elif opcode == OpCode.MASK_IF_DICT:
                if not isinstance(self.focus_stack[-1], dict):
                    self.backtrack()
                else:
                    self.focus_stack[-1] = MaskedDict(self.focus_stack[-1])

            elif opcode == OpCode.FOCUS_ON_HEAD:
                if len(self.focus_stack[-1]) == 0:
                    self.backtrack()
                else:
                    self.focus_stack.append(self.focus_stack[-1].pop())

            elif opcode == OpCode.FOCUS_ON_KEY:
                if instruction[1] not in self.focus_stack[-1]:
                    self.backtrack()
                else:
                    self.focus_stack.append(self.focus_stack[-1].pop(instruction[1]))

            elif opcode == OpCode.FOCUS_ON_BINDING:
                self.focus_stack.append(self.bindings[instruction[1]])

            elif opcode == OpCode.POP_FOCUS:
                self.focus_stack.pop()

            elif opcode == OpCode.YIELD_BINDINGS:
                yield {v: freeze(b) for v, b in self.bindings.items()}
                self.backtrack()

            else:
                raise JertlInterpreterException(f'OpCode {opcode} not understood')



import collections.abc
import copy
from   functools import singledispatch


class MaskedList(collections.abc.Sequence):
    def __init__(self, backing_sequence, offset=0, end=None):
        self.backing_sequence = backing_sequence
        self.offset = offset
        self.end = end if end is not None else len(backing_sequence)
        if not 0 <= self.end <= len(backing_sequence):
            raise IndexError(self.end)

    def __len__(self):
        return self.end - self.offset

    def __getitem__(self, index):
        offset_index = index + self.offset
        if offset_index < self.end:
            return self.backing_sequence[offset_index]
        else:
            raise IndexError(index)

    def __eq__(self, other):
        return other == self.backing_sequence[self.offset:self.end]

    def pop(self):
        value = self[0]
        self.offset += 1
        return value

    def widen(self):
        if self.end < len(self.backing_sequence):
            self.end += 1
        else:
            raise IndexError(len(self))

    def collapse(self):
        self.end = self.offset

    def snapshot(self):
        return MaskedList(self.backing_sequence, self.offset, self.end)

    def freeze(self):
        return [*self]

    def __repr__(self):
        return repr(self.backing_sequence[self.offset:self.end])


class MaskedDict(collections.abc.Mapping):
    def __init__(self, backing_mapping, masked_keys=None):
        self.backing_mapping = backing_mapping
        self.masked_keys = set() if masked_keys is None else copy.copy(masked_keys)

    def __len__(self):
        return len(self.backing_mapping) - len(self.masked_keys)

    def __getitem__(self, key):
        if key in self.masked_keys:
            raise KeyError(key)
        else:
            return self.backing_mapping[key]

    def __delitem__(self, key):
        if key not in self.backing_mapping or key in self.masked_keys:
            raise KeyError(key)
        else:
            self.masked_keys |= {key}

    def pop(self, key):
        value = self[key]
        del self[key]
        return value

    def snapshot(self):
        return MaskedDict(self.backing_mapping, self.masked_keys)

    def freeze(self):
        return {**self}

    class Iterator:
        def __init__(self, masked_mapping):
            self.masked_keys    = masked_mapping.masked_keys
            self.inner_iterator = iter(masked_mapping.backing_mapping)

        def __next__(self):
            candidate_next = next(self.inner_iterator)
            if candidate_next in self.masked_keys:
                return self.__next__()
            else:
                return candidate_next

    def __iter__(self):
        return self.Iterator(self)

    def __repr__(self):
        return repr({k: v for k, v in self.backing_mapping.items()
                          if k not in self.masked_keys})


@singledispatch
def freeze(value):
    return value

@freeze.register(MaskedList)
def _(masked_list):
    return masked_list.freeze()

@freeze.register(MaskedDict)
def _(masked_dict):
    return masked_dict.freeze()


@singledispatch
def snapshot(value):
    return value

@snapshot.register(MaskedList)
def _(masked_list):
    return masked_list.snapshot()

@snapshot.register(MaskedDict)
def _(masked_dict):
    return masked_dict.snapshot()



import jertl.ast.emitter as jae
import jertl.ast.builder as jab

from jertl.engine.interpreter import Interpreter
from jertl.engine.construct   import construct

from jertl.results import Match, Transformation, Collation, Inference

class Matcher:
    """Class which compares data to a structure and returns all Matches"""
    def __init__(self, structure):
        """
        Args:
            structure (str): pattern describing structure to be identified
        """
        self._ast          = jab.ast_for_string(structure, 'toplevel_structure')
        self._instructions = list(jae.emit_match(self._ast))

    def match_all(self, data):
        """match_all Generate all Matches of a structure to data.

        Args:
            data ((Sequence | Mapping | Number)): Python data structure
                to examine

        Yields:
            :obj:`~.Match`: All possible Matches

        """
        interpreter = Interpreter(self._instructions)
        for bindings in interpreter.match_all(data):
            yield Match(self._ast, bindings)

    def match(self, data):
        """match Matches a structure to data.

        Args:
            data ((Sequence | Mapping | Number)): Python data structure to match

        Returns:
            Optional(:obj:`~.Match`): (Sequence | Mapping | Number)
        """
        for m in self.match_all(data):
            return m


class Filler:
    """Class which performs fills"""
    def __init__(self, structure):
        """
        Args:
            structure (str): pattern describing structure to be used for creating new data
        """
        self._ast = jab.ast_for_string(structure, 'toplevel_structure')

    def fill(self, **bindings):
        """fill Fills structure and returns data with structure variables replaced with their bindings

        Args:
            **bindings (Dict[str, list | dict | str | Number]): Initial variable bindings

        Returns:
            (list | dict | str | Number): data
        """
        return construct(self._ast, bindings)


class Transformer:
    """A transformer with compile structure"""
    def __init__(self, transform):
        """
        Args:
            transform (str): pattern describing a transform
        """
        self._ast = jab.ast_for_string(transform, 'transform')
        self._source = self._ast.input
        self._target = self._ast.output
        self._instructions = list(jae.emit_match(self._source))

    def transform_all(self, data):
        """transform_all finds all possible transforms of data

        Args:
            data ((list | dict | str | Number)): Data structure to be
                transformed

        Yields:
            :obj:`~.Transformation`:  All possible transforms of the sources

        """
        interpreter = Interpreter(self._instructions)
        for bindings in interpreter.match_all(data):
            yield Transformation(self._source, self._target, bindings)

    def transform(self, data):
        """transform Finds a transform of data

        Args:
            data ((list | dict | str | Number)): Data structure to be
                transformed

        Returns:
            Optional(:obj:`~.Transformer`):: A Transformation if one was found

        """
        for m in self.transform_all(data):
            return m


class Collator:
    """Class which performs collations"""
    def __init__(self, collation):
        """
        Args:
            collation (str): pattern describing a collation

        """
        self._ast = jab.ast_for_string(collation, 'collation')
        self._toplevel = {m.variable.identifier for m in self._ast}
        self._instructions = list(jae.emit_collation(self._ast, self._toplevel))

    def collate_all(self, **bindings):
        """collate_all Yield all possible collations of data

        Args:
            **bindings (Dict[str, list | dict | str | Number]): Initial variable bindings

        Yields:
            :obj:`~.Collation`: All possible collations

        """
        interpreter = Interpreter(self._instructions)
        for bindings in interpreter.match_all(None, bindings):
            structures = {m.variable.identifier: m.structure for m in self._ast}
            yield Collation(structures, bindings)

    def collate(self, **bindings):
        """collate Find a collation if any

        Args:
            **bindings (Dict[str, list | dict | str | Number]): Initial variable bindings

        Returns:
            Optional(:obj:`Collation`): A Collation

        """
        for m in self.collate_all(**bindings):
            return m


class Rule:
    """Class which finds inferences of a production rule"""
    def __init__(self, rule):
        """
        Args:
            rule (str): pattern describing an inference rule
        """
        self._ast = jab.ast_for_string(rule, 'rule_')
        self._toplevel = {m.variable.identifier for m in self._ast.matchers}
        self._instructions = list(jae.emit_collation(self._ast.matchers, self._toplevel))

    def infer_all(self, **bindings):
        """infer_all Yield all possible inferences of rule applied to data

        Args:
            **bindings (Dict[str, list | dict | str | Number]): Initial variable bindings

        Yields:
            :obj:`~.Inference`: All possible inferences

        """
        interpreter = Interpreter(self._instructions)
        for bindings in interpreter.match_all(None, bindings):
            inputs  = {m.variable.identifier: m.structure for m in self._ast.matchers}
            outputs = {a.variable.identifier: a.structure for a in self._ast.setters}
            yield Inference(inputs, outputs, bindings)

    def infer(self, **bindings):
        """infer Finds an inference of rule applied to data (if any)

        Args:
            **bindings (Dict[str, list | dict | str | Number]): Initial variable bindings

        Returns:
            Optional(:obj:`~.Inference`): An inference

        """
        for m in self.infer_all(**bindings):
            return m


__all__ = ['Matcher', 'Filler', 'Transformer', 'Collator', 'Rule']



import antlr4
from antlr4.error.Errors import ParseCancellationException

from jertl.exceptions import JertlSyntaxError

from .  import representation as jar
from .  import helpers        as helpers
from .. import parser         as jp

class JertlBuilder(helpers.ASTBuilder, jp.Listener):

    def exitTransform(self, ctx):
        ctx.AST = jar.Transform(input=ctx.structure(0).AST,
                                output=ctx.structure(1).AST)

    def exitCollation(self, ctx):
        ctx.AST = self.ASTsOf(ctx.matcher())

    def exitRule_(self, ctx):
        ctx.AST = jar.Rule(matchers=self.ASTsOf(ctx.matcher()),
                           setters=self.ASTsOf(ctx.setter()))

    def exitMatcher(self, ctx):
        ctx.AST = jar.Matcher(ctx.variable().AST, ctx.structure().AST)

    def exitSetter(self, ctx):
        ctx.AST = jar.Setter(ctx.variable().AST, ctx.structure().AST)

    def exitStructure(self, ctx):
        ctx.AST = self.whichOf(ctx.obj(),
                               ctx.array(),
                               ctx.atom(),
                               ctx.variable()).AST

    def exitObj(self, ctx):
        ast = dict(self.ASTsOf(ctx.key_values))
        if ctx.kwargs() is not None:
            ast[jar.KWArgs()] = ctx.kwargs().AST
        ctx.AST = ast

    def exitKey_value(self, ctx):
        ctx.AST = (self.dequote(ctx.STRING().getText()), ctx.structure().AST)

    def exitKwargs(self, ctx):
        ctx.AST = ctx.variable().AST

    def exitArray(self, ctx):
        ctx.AST = self.ASTsOf(ctx.elements)

    def exitElement(self, ctx):
        ctx.AST = self.whichOf(ctx.structure(), ctx.varargs()).AST

    def exitVarargs(self, ctx):
        ctx.AST = jar.VarArgs(ctx.variable().AST)

    def exitAtom(self, ctx):
        if ctx.NULL() is not None:
            ctx.AST = None
        elif ctx.TRUE() is not None:
            ctx.AST = True
        elif ctx.FALSE() is not None:
            ctx.AST = False
        elif ctx.INTEGER() is not None:
            ctx.AST = int(ctx.INTEGER().getText())
        elif ctx.FLOAT() is not None:
            ctx.AST = float(ctx.FLOAT().getText())
        elif ctx.STRING() is not None:
            ctx.AST = self.dequote(ctx.STRING().getText())

    def exitVariable(self, ctx):
        ctx.AST = jar.Variable(ctx.IDENTIFIER().getText())

    def exitToplevel_structure(self, ctx):
        ctx.AST = ctx.structure().AST

    def dequote(self, s):
        if s.endswith('"'):
            s = s[:-1]
        if s.startswith('"'):
            s = s[1:]
        return s

def string_parser(string):
    input_stream = antlr4.InputStream(string)
    lexer = jp.Lexer(input_stream)
    lexer.removeErrorListeners()
    lexer.addErrorListener(helpers.ThrowingErrorListener())

    token_stream = antlr4.CommonTokenStream(lexer)

    parser = jp.Parser(token_stream)
    parser.removeErrorListeners()
    parser.addErrorListener(helpers.ThrowingErrorListener())
    #
    return parser

def parse_string(string, rulename):
    parser = string_parser(string)
    rule   = getattr(parser, rulename)
    tree   = rule()
    return tree

def ast_from_tree(tree, trace=False):
    builder = JertlBuilder(trace=trace)
    walker  = antlr4.ParseTreeWalker()
    walker.walk(builder, tree)
    return tree.AST

def ast_for_string(string, rulename, trace=False):
    return ast_from_tree(parse_string(string, rulename), trace=trace)



# SPDX-FileCopyrightText: 2022-present RaymondPelletier <68929475+RaymondPelletier@users.noreply.github.com>
# SPDX-License-Identifier: Apache-2.0

from .processors import Matcher, Filler, Transformer, Collator, Rule

def match(structure, data):
    """Matches a structure to data.

    Args:
        structure (str): A pattern defining the structure to compare to
            data.
        data ((Sequence | Mapping | Number)): Python data to compare
            structure against

    Returns:
        Optional(:obj:`~.Match`): A description of the match

    """
    return Matcher(structure).match(data)

def match_all(structure, data):
    """Generate all Matches of a structure to data.

    Args:
        structure (str): A pattern defining the structure to compare to
            the target data.
        data ((Sequence | Mapping | Number)): Python data to match

    Yields:
        :obj:`~.Match`: All possible matches

    """
    yield from Matcher(structure).match_all(data)

def compile_match(structure):
    """Compiles a structure pattern for later use.

    Args:
        structure (str): A pattern defining the structure to compare to
            the target data.

    Returns:
        :obj:`~.Matcher`: An object which can be used to match template to objects

    """
    return Matcher(structure)

def fill(structure, **bindings):
    """Populate a structure.

    Args:
        structure (str): A pattern defining the structure to compare to
            the target data.
        **bindings (Dict[str, list | dict | str | Number]): values for free variables referenced by
            template

    Returns:
        Dict | Sequence | Number: A filled template

    """
    return Filler(structure).fill(**bindings)

def compile_fill(structure):
    """Compiles a structure for later filling.

    Args:
        structure (str): A pattern defining the structure to compare to
            the target data.

    Returns:
        :obj:`~.Filler`: A filler
    """
    return Filler(structure)

def transform(transform, data):
    """Matches a structure to data and fills a target structure.

    Args:
        transform (str): A pattern describing structures for matching
            and filling.
        data (Dict | Sequence | Number): Data to match against.

    Returns:
        Optional(:obj:`~.Transformer`):: A Transformation
    """
    return Transformer(transform).transform(data)

def transform_all(transform, data):
    """Generate all Transforms of data given Transformation specification.

    Args:
        transform (str): A pattern describing structures for matching
            and filling.
        data (KwArgs): Data to match against.

    Yields:
        :obj:`~.Transformation`:  All possible transforms of the sources

    """
    yield from Transformer(transform).transform_all(data)

def compile_transform(transform):
    """Compiles a transform pattern for later reuse.

    Args:
        transform (str): Pattern describing a transform.

    Returns:
        :obj:`~.Transformer`: A transformer
    """
    return Transformer(transform)


def collate(collation, **bindings):
    """Match all matchers to targets

    Args:
        collation (str): Specification of collation.
        **bindings (Dict[str, list | dict | str | Number]): Data to match against.

    Returns:
        Optional(:obj:`Collation`): A Collation
    """
    return Collator(collation).collate(**bindings)

def collate_all(collation, **sources):
    """Yields all matches to targets

    Args:
        collation (str): Specification of collation.
        **sources (Dict[str, list | dict | str | Number]): Data to match against.

    Yields:
        :obj:`~.Collation`: All possible collations

    """
    yield from Collator(collation).collate_all(**sources)

def compile_collate(collation):
    """Compiles a template for later reuse.

    Args:
        collation (str): Specification of collation.
        sources (Dict[str, list | dict | str | Number]): Data to match against.

    Returns:
        :obj:`~.Collator`: A collator
    """
    return Collator(collation)

def infer(rule, **sources):
    """Applies a production rule agains sources and generates new data according to output structures

    Args:
        rule (str): Pattern describing a production rule.
        **sources (Dict[str, list | dict | str | Number]): Data to match against.

    Returns:
        Optional(:obj:`~.Inference`): An inference
    """
    return Rule(rule).infer(**sources)

def infer_all(rule, **sources):
    """Generate all application of rule.

    Args:
        rule (str): Pattern describing a production rule.
        **sources (Dict[str, list | dict | str | Number]): Data to match against.

    Yields:
        :obj:`~.Inference`: All possible inferences

    """
    yield from Rule(rule).infer_all(**sources)

def compile_rule(rule):
    """Compiles a inference rule for later reuse.

    Args:
        rule (str): Pattern describing a production rule.

    Returns:
        :obj:`~.Rule`: A rule
    """
    return Rule(rule)

__all__ = ['match',     'match_all',     'compile_match',
           'fill',                       'compile_fill',
           'transform', 'transform_all', 'compile_transform',
           'collate',   'collate_all',   'compile_collate',
           'infer',     'infer_all',     'compile_rule']



import pytest

from jertl.ast.builder        import ast_for_string
from jertl.engine.construct   import construct, ensure_purity


tests = [['x',
          {'x': 100},
          100],

         ['[x, y, "z"]',
           {'x': True, 'y': None},
           [True, None, 'z']],

         ['[1, 2, *splat]',
           {'splat': [3, 4, 5]},
           [1, 2, 3, 4, 5]],

         ['[1, 2.2, true, null, "String"]',
           {},
           [1, 2.2, True, None, 'String']],

         ['{"1": 1, "2": 2, **splot}',
          {'splot': {"3": 3, "4": 4, "5": 5}},
          {'1': 1, '2': 2, '3': 3, '4': 4, '5': 5}],

         ['{"1": 1, "list": [1, {"inner": x}, *splat, y], **splot}',
            {'x': True, 'y': None, 'splat': [3, 4, 5], 'splot': {"3": 3, "4": 4, "5": 5}},
            {'1': 1, 'list': [1, {'inner': True}, 3, 4, 5, None], '3': 3, '4': 4, '5': 5}]]

@pytest.mark.parametrize('pattern,bindings,result', tests)
def test_ast_for_string(pattern, bindings, result):
    structure = ast_for_string(pattern, 'toplevel_structure')
    constructed = construct(structure, bindings)
    assert constructed == result
    ensure_purity(constructed)



import pytest
from jertl.ast.builder import parse_string
from jertl.exceptions  import JertlSyntaxError

positive_tests = [['structure', '{"an": "object"}'],
                  ['structure', '["a", 4, "element", array]'],
                  ['structure', '"atom"'],
                  ['structure', 'variable'],

                  ['array',     '[1, 2, True]'],
                  ['array',     '[]'],
                  ['array',     '[1, "have a", *splat, "or", *two]'],
                  ['atom',      'null'],
                  ['atom',      'true'],
                  ['atom',      'false'],
                  ['atom',      '1'],
                  ['atom',      '3.14159'],
                  ['atom',      '"string"'],
                  ['obj',       '{"integer": 1, "list": [a, 4.4, *spread], **kw}'],

                  ['toplevel_structure', '{"an": "object"}'],
                  ['toplevel_structure', '["a", 4, "element", array]'],
                  ['toplevel_structure', '"atom"'],
                  ['toplevel_structure', 'variable'],

                  ['transform', '{"integer": 1, "list": [a, 4.4, *spread], **kw} --> True'],
                  ['matcher',   'thing ~ {"integer": 1, "list": [a, 4.4, *spread], **kw}'],
                  ['setter',    'result := {"integer": 1}'],
                  ['setter',    'result := {"integer": 1, "list": [a, 4.4, *spread], **kw}'],
                  ['collation', '''
                                    thing1 ~ {"integer": 1, "list": [_, 4.4, *spread], **kw}
                                    thing2 ~ thing1
                                    thing3 ~ [thing1, thing2]
                                '''],
                  ['rule_',     '''
                                    thing1 ~ {"integer": 1, "list": [a, 4.4, *spread], **kw}
                                    thing2 ~ thing1
                                    thing3 ~ [thing1, thing2]
                                  -->
                                    thong1 := {"integer": 2, "list": [thing3, 4.4, *spread], **kw}
                                ''']]

@pytest.mark.parametrize('rule,source', positive_tests)
def test_parse(rule, source):
    parse_string(source, rule)

def test_failure():
    with pytest.raises(JertlSyntaxError):
        parse_string('[$internal_variable_name]', 'structure')

def test_null_collation():
    parse_string('', 'collation')

def test_null_rule():
    parse_string('-->', 'rule_')

negative_tests = [['toplevel_structure', '{"an": "object"}_'],
                  ['toplevel_structure', '["a", 4, "element", array]['],
                  ['toplevel_structure', '\'atom\''],
                  ['toplevel_structure', 'var iable'],

                  ['transform', '{"integer": 1, "list": [a, 4.4, *spread], **kw} --> True;'],
                  ['collation', '''
                                    thing1 ~ {"integer": 1, "list": [_, 4.4, *spread], **kw}
                                    thing2 ~ thing1
                                    thing3 ~ [thing1, thing2] -->
                                '''],
                  ['rule_',     '''
                                    thing1 ~ {"integer": 1, "list": [a, 4.4, *spread], **kw}
                                    thing2 ~ thing1
                                    thing3 ~ [thing1, thing2]
                                  -->
                                    thong1 := {"integer": 2, "list": [thing3, 4.4, *spread], **kw}
                                    ()
                                ''']]


@pytest.mark.parametrize('rule,source', negative_tests)
def test_invalid_input(rule, source):
    with pytest.raises(JertlSyntaxError):
        parse_string(source, rule)



import antlr4

from antlr4.error.ErrorListener import ErrorListener
from ..exceptions import JertlSyntaxError

class ThrowingErrorListener(ErrorListener):
    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        ex = JertlSyntaxError(f'line {line}: {column} {msg}')
        ex.line = line
        ex.column = column
        raise ex

def remove_suffix(string, suffix):
    if string.endswith(suffix):
        return string[:-len(suffix)]
    else:
        return string

class TracingListener(antlr4.ParseTreeListener):
    def __init__(self, *varargs, trace=False, **kwargs):
        self._level = 0
        self._trace = trace
        super().__init__(*varargs, **kwargs)

    def context2name(self, ctx):
        return remove_suffix(type(ctx).__name__.split('.')[-1], 'Context')

    def enterEveryRule(self, ctx):
        if self._trace:
            print(' '*self._level, '-->', self.context2name(ctx))
            self._level += 1
        #
        super().enterEveryRule(ctx)

    def exitEveryRule(self, ctx):
        name = self.context2name(ctx)
        if self._trace:
            self._level -= 1
            print(' '*self._level, '<--', name)
        #
        super().exitEveryRule(ctx)

# For use by Listeners generating ASTs
class ASTBuilder(TracingListener):
    def __init__(self, *varargs, ignore_missing=False, **kwargs):
        self.ignore_missing = ignore_missing
        super().__init__(*varargs, **kwargs)

    def exitEveryRule(self, ctx):
        self.checkHasAst(ctx)
        super().exitEveryRule(ctx)

    def checkHasAst(self, ctx):
        if not hasattr(ctx, 'AST') and not self.ignore_missing:
            raise Exception(self.context2name(ctx), 'has no AST')
        else:
            return True

    def _AST(self, context):
        if self.ignore_missing:
            return getattr(context, 'AST', None)
        else:
            return context.AST

    def ASTsOf(self, contexts):
        return [self._AST(c) for c in contexts]

    def whichOf(self, *contexts):
        for ctx in contexts:
            if ctx is not None:
                return ctx

    def optional(self, ctx):
        if ctx is not None:
            return self._AST(ctx)



import jertl
from   jertl.engine.construct import ensure_purity

def test_match_all():
    matches = list(jertl.match_all('[*x, [y, [*z, *z], y], *x]', [1, [2, [3, 3], 2] ,1]))
    assert len(matches) == 1
    assert matches[0].bindings['x'] == [1]
    assert matches[0].bindings['y'] ==  2
    assert matches[0].bindings['z'] == [3]

def test_compiled_match():
    matcher = jertl.compile_match('[*before, x, *after]')
    for i, match in enumerate(matcher.match_all(list(range(3)))):
        assert match.bindings['before'] == list(range(i))
        assert match.bindings['x']      == i
        assert match.bindings['after']  == list(range(i+1, 3))

        ensure_purity(match.matched)

def test_fill():
    assert jertl.fill('[a, b, c]', a=True, b=None, c='a string') == [True, None, 'a string']

def test_compiled_fill():
    filler = jertl.compile_fill('[a, b]')

    for pair in [[1,2], [7,9]]:
      assert filler.fill(a=pair[0], b=pair[1]) == pair

def test_collate_all():
    collations = list(jertl.collate_all('''
                                      employee   ~ {"Name": "Ray", "Supervisor": supervisor_id}
                                      supervisor ~ {"EmployeeId": supervisor_id}
                                      ''',
                                      employee={'Name': 'Ray', 'Supervisor': 666},
                                      supervisor={'EmployeeId': 666}))
    assert len(collations) == 1
    assert collations[0].bindings == {'employee': {'Name': 'Ray', 'Supervisor': 666}, 'supervisor': {'EmployeeId': 666}, 'supervisor_id': 666}

def test_compiled_collate():
    collator = jertl.compile_collate('''
                                   a ~ [*_, n, *_]
                                   b ~ [*_, n, *_]
                                   ''')
    collations = list(collator.collate_all(a=[1, 2, 3, 4], b=[4, 2]))
    assert len(collations) == 2
    assert collations[0].bindings['n'] == 2
    assert collations[1].bindings['n'] == 4

def test_transform_all():
    transforms = list(jertl.transform_all('[*a, x, b, y, *c] --> [y, x]', [1,2,3,4,5]))
    assert len(transforms) == 3
    assert transforms[0].filled == [3, 1]
    assert transforms[1].filled == [4, 2]
    assert transforms[2].filled == [5, 3]

def test_compiled_inference():
    rule = jertl.compile_rule('''
							a ~ [*_1, n, x, *_2]
							b ~ [*_3, n, y, *_4]
						-->
							c := [n, [x, y]]
					    ''')

    inferences = list(rule.infer_all(a=[1, 2, 3, 4, 12], b=[4, 2, 7]))
    assert len(inferences) == 2
    assert inferences[0].fills == {'c': [2, [3, 7]]}
    assert inferences[1].fills == {'c': [4, [12, 2]]}



import pytest
from jertl.ast.builder        import ast_for_string
from jertl.ast.representation import Variable, VarArgs, KWArgs
from jertl.ast.representation import Matcher,  Setter,  Transform, Rule

tests = [['structure', '{"an": "object"}',
                        {'an': 'object'}],

         ['structure', '["a", 4, "element", array]',
                        ['a', 4, 'element', Variable('array')]],

         ['structure', '"atom"',
                        'atom'],

         ['structure', 'variable',
                        Variable('variable')],

         ['array',     '[1, 2, true]',
                        [1, 2, True]],

         ['array',     '[]',
                        []],

         ['array',     '[1, "have a", *splat,                     "or", *two]',
                        [1, 'have a', VarArgs(Variable('splat')), 'or', VarArgs(Variable('two'))]],

         ['atom',      'null',
                        None],

         ['atom',      'true',
                        True],

         ['atom',      'false',
                        False],

         ['atom',      '1',
                        1],

         ['atom',      '3.14159',
                        3.14159],

         ['atom',      '"string"',
                        'string'],

         ['obj',       '{"integer": 1, "list": [a,             4.4, *spread],                  **kw}',
                        {'integer': 1,
                         'list': [Variable('a'), 4.4, VarArgs(variable=Variable('spread'))],
                         KWArgs(): Variable('kw')}],

         ['transform', '{"integer": 1, "list": [a, 4.4, *spread], **kw} --> true',
                        Transform(input={'integer': 1,
                                         'list': [Variable('a'), 4.4, VarArgs(variable=Variable('spread'))],
                                         KWArgs(): Variable('kw')},
                                  output=True)],

         ['matcher',   'thing ~ {"integer": 1, "list": [a, 4.4, *spread], **kw}',
                        Matcher(variable=Variable('thing'),
                                structure={'integer': 1,
                                           'list': [Variable('a'), 4.4, VarArgs(variable=Variable('spread'))],
                                           KWArgs(): Variable('kw')})],

         ['setter',    'result := {"integer": 1}',
                        Setter(variable=Variable('result'), structure={'integer': 1})],

         ['setter',    'result := {"integer": 1, "list": [a, 4.4, *spread], **kw}',
                        Setter(variable=Variable('result'),
                               structure={'integer': 1,
                                          'list':    [Variable('a'),
                                                      4.4,
                                                      VarArgs(variable=Variable('spread'))],
                                          KWArgs():  Variable('kw')})],

         ['collation', '''
                         thing1 ~ {"integer": 1, "list": [x, 4.4, *spread], **kw}
                         thing2 ~ thing1
                         thing3 ~ [thing1, thing2]
                     ''',
                     [Matcher(variable=Variable('thing1'),
                              structure={'integer': 1,
                                         'list': [Variable('x'), 4.4, VarArgs(variable=Variable('spread'))],
                                          KWArgs():  Variable('kw')}),
                      Matcher(variable=Variable('thing2'), structure=Variable('thing1')),
                      Matcher(variable=Variable('thing3'), structure=[Variable('thing1'), Variable('thing2')])]],

         ['rule_',     '''
                           thing1 ~ {"integer": 1, "list": [a, 4.4, *spread], **kw}
                           thing2 ~ thing1
                           thing3 ~ [thing1, thing2]
                         -->
                           thong1 := {"integer": 2, "list": [thing3, 4.4, *spread], **kw}
                       ''',
                     Rule(matchers=[Matcher(variable=Variable('thing1'),
                                            structure={KWArgs(): Variable('kw'),
                                                        'integer': 1,
                                                        'list': [Variable('a'), 4.4, VarArgs(variable=Variable('spread'))]}),
                                    Matcher(variable=Variable('thing2'), structure=Variable('thing1')),
                                    Matcher(variable=Variable('thing3'), structure=[Variable('thing1'), Variable('thing2')])],
                          setters=[Setter(variable=Variable('thong1'),
                                          structure={'integer': 2,
                                                     'list': [Variable('thing3'),
                                                              4.4,
                                                              VarArgs(variable=Variable('spread'))],
                                                     KWArgs(): Variable('kw')})])],

         ['rule_',     '-->',  Rule(matchers=[], setters=[])],
         ['collation', '',     []]]

@pytest.mark.parametrize('rule,source,result', tests)
def test_ast_for_string(rule, source, result):
    ast = ast_for_string(source, rule)
    assert ast == result



from jertl.ast.representation import Variable

# Variables

def test_equality():
    assert Variable('x') == Variable('x')

def test_inequality():
    assert Variable('x') != Variable('y')

def test_uniqueness_of_anonymous_variables():
    assert Variable('_') != Variable('_')

def test_representation_of_anonymous_variables():
    assert repr(Variable('_')) == '_'

def test_identifier_of_anonymous_variables():
    anonymous_variable = Variable('_')
    assert anonymous_variable.identifier != '_'
    assert anonymous_variable.identifier.startswith('$')



from dataclasses import dataclass, field
from typing      import Dict, List, Union

@dataclass
class Variable:
    identifier: str
    anonymous:  bool = field(init=False, default=False)
    #
    anon_counter = 0

    def __post_init__(self):
        if self.identifier == '_':
            Variable.anon_counter += 1
            self.anonymous = True
            self.identifier = f'$_{Variable.anon_counter}'

    def __repr__(self):
        return self.identifier if not self.anonymous else '_'

@dataclass
class VarArgs:
    variable: Variable

## Used by emitter to identify KWArgs in a dictionary
class KWArgs:
    def __repr__(self):
        return '<KWArgs>'

    ## Equality and hash used by pytest cases
    def __eq__(self, other):
        return type(other) == KWArgs

    def __hash__(self):
        return hash(repr(self))

Structure = Union[List, Dict, str, int, float, bool, type(None)]

@dataclass
class Transform:
    input:  Structure
    output: Structure

@dataclass
class Matcher:
    variable:  Variable
    structure: Structure

@dataclass
class Setter:
    variable:  Variable
    structure: Structure

@dataclass
class Rule:
    matchers: List[Matcher]
    setters:  List[Setter]



from . import JertlParser
from . import JertlLexer
from . import JertlListener

class Parser(JertlParser.JertlParser):
    pass

class Lexer(JertlLexer.JertlLexer):
    pass

class Listener(JertlListener.JertlListener):
    pass



# Generated from Jertl.g4 by ANTLR 4.11.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .JertlParser import JertlParser
else:
    from JertlParser import JertlParser

# This class defines a complete listener for a parse tree produced by JertlParser.
class JertlListener(ParseTreeListener):

    # Enter a parse tree produced by JertlParser#toplevel_structure.
    def enterToplevel_structure(self, ctx:JertlParser.Toplevel_structureContext):
        pass

    # Exit a parse tree produced by JertlParser#toplevel_structure.
    def exitToplevel_structure(self, ctx:JertlParser.Toplevel_structureContext):
        pass


    # Enter a parse tree produced by JertlParser#transform.
    def enterTransform(self, ctx:JertlParser.TransformContext):
        pass

    # Exit a parse tree produced by JertlParser#transform.
    def exitTransform(self, ctx:JertlParser.TransformContext):
        pass


    # Enter a parse tree produced by JertlParser#collation.
    def enterCollation(self, ctx:JertlParser.CollationContext):
        pass

    # Exit a parse tree produced by JertlParser#collation.
    def exitCollation(self, ctx:JertlParser.CollationContext):
        pass


    # Enter a parse tree produced by JertlParser#rule_.
    def enterRule_(self, ctx:JertlParser.Rule_Context):
        pass

    # Exit a parse tree produced by JertlParser#rule_.
    def exitRule_(self, ctx:JertlParser.Rule_Context):
        pass


    # Enter a parse tree produced by JertlParser#matcher.
    def enterMatcher(self, ctx:JertlParser.MatcherContext):
        pass

    # Exit a parse tree produced by JertlParser#matcher.
    def exitMatcher(self, ctx:JertlParser.MatcherContext):
        pass


    # Enter a parse tree produced by JertlParser#setter.
    def enterSetter(self, ctx:JertlParser.SetterContext):
        pass

    # Exit a parse tree produced by JertlParser#setter.
    def exitSetter(self, ctx:JertlParser.SetterContext):
        pass


    # Enter a parse tree produced by JertlParser#structure.
    def enterStructure(self, ctx:JertlParser.StructureContext):
        pass

    # Exit a parse tree produced by JertlParser#structure.
    def exitStructure(self, ctx:JertlParser.StructureContext):
        pass


    # Enter a parse tree produced by JertlParser#obj.
    def enterObj(self, ctx:JertlParser.ObjContext):
        pass

    # Exit a parse tree produced by JertlParser#obj.
    def exitObj(self, ctx:JertlParser.ObjContext):
        pass


    # Enter a parse tree produced by JertlParser#key_value.
    def enterKey_value(self, ctx:JertlParser.Key_valueContext):
        pass

    # Exit a parse tree produced by JertlParser#key_value.
    def exitKey_value(self, ctx:JertlParser.Key_valueContext):
        pass


    # Enter a parse tree produced by JertlParser#kwargs.
    def enterKwargs(self, ctx:JertlParser.KwargsContext):
        pass

    # Exit a parse tree produced by JertlParser#kwargs.
    def exitKwargs(self, ctx:JertlParser.KwargsContext):
        pass


    # Enter a parse tree produced by JertlParser#array.
    def enterArray(self, ctx:JertlParser.ArrayContext):
        pass

    # Exit a parse tree produced by JertlParser#array.
    def exitArray(self, ctx:JertlParser.ArrayContext):
        pass


    # Enter a parse tree produced by JertlParser#element.
    def enterElement(self, ctx:JertlParser.ElementContext):
        pass

    # Exit a parse tree produced by JertlParser#element.
    def exitElement(self, ctx:JertlParser.ElementContext):
        pass


    # Enter a parse tree produced by JertlParser#varargs.
    def enterVarargs(self, ctx:JertlParser.VarargsContext):
        pass

    # Exit a parse tree produced by JertlParser#varargs.
    def exitVarargs(self, ctx:JertlParser.VarargsContext):
        pass


    # Enter a parse tree produced by JertlParser#atom.
    def enterAtom(self, ctx:JertlParser.AtomContext):
        pass

    # Exit a parse tree produced by JertlParser#atom.
    def exitAtom(self, ctx:JertlParser.AtomContext):
        pass


    # Enter a parse tree produced by JertlParser#variable.
    def enterVariable(self, ctx:JertlParser.VariableContext):
        pass

    # Exit a parse tree produced by JertlParser#variable.
    def exitVariable(self, ctx:JertlParser.VariableContext):
        pass



del JertlParser


from jertl.engine.construct import construct

class Bindings:
    """Base class for jertl processing results"""
    def __init__(self, bindings):
        self._bindings = bindings

    @property
    def bindings(self):
        """
        bindings A dictionary mapping variable identifiers to their bindings

        Returns:
            Dict[str, list | dict | str | Number]: dictionary
                containing a set of variable bindings
        """
        return {k: v for k, v in self._bindings.items() if not k.startswith('$')}


class Match(Bindings):
    """Output of a matcher"""
    def __init__(self, matched, bindings):
        self._matched = matched
        super().__init__(bindings)

    @property
    def matched(self):
        """
        matched The result of filling the structure pattern with the variables bound during the match process

        Returns:
            list | dict | str | Number: Result of a fill using the structure pattern and variable bindings
        """
        return construct(self._matched, self._bindings)


class Transformation(Bindings):
    """Class describing a tranform result"""
    def __init__(self, matched, filled, bindings):
        self._matched = matched
        self._filled  = filled
        super().__init__(bindings)

    @property
    def matched(self):
        """
        matched matching structure with variables replace with their bindings

        Returns:
            list | dict | str | Number: Result of a fill using the matching structure pattern and variable bindings
        """
        return construct(self._matched, self._bindings)

    @property
    def filled(self):
        """
        filled transformed structure

        Returns:
            list | dict | str | Number: Result of a fill using the target structure pattern and variable bindings
        """
        return construct(self._filled, self._bindings)


class Collation(Bindings):
    """Output of a collator"""
    def __init__(self, matchers, bindings):
        self._matchers = matchers
        super().__init__(bindings)

    @property
    def matches(self):
        """
        matches The result of filling the targetted match structures with the variables bound during the match process

        Returns:
            Dict[str, list | dict | str | Number]: Mapping of target identifiers to filled matching structures
        """
        return {v: construct(t, self._bindings) for v, t in self._matchers.items()}


class Inference(Bindings):
    """Result yielded by an Inferencer"""
    def __init__(self, matchers, fillers, bindings):
        self._matchers = matchers
        self._fillers  = fillers
        super().__init__(bindings)

    @property
    def matches(self):
        """
        result The result of filling the targetted match structures with the variables bound during the match process

        Returns:
            Dict[str, list | dict | str | Number]: Mapping of variable identifiers to structures
        """
        return {v: construct(t, self._bindings) for v, t in self._matchers.items()}

    @property
    def fills(self):
        """
        result The result of filling the targetting fill structures with the variables bound during the match process

        Returns:
            Dict[str, list | dict | str | Number]: Mapping of variable identifiers to structures
        """
        return {v: construct(t, self._bindings) for v, t in self._fillers.items()}



import enum

class OpCode(enum.Enum):
    MATCH_VALUE      = enum.auto()
    BIND_VARIABLE    = enum.auto()
    MATCH_VARIABLE   = enum.auto()

    BIND_VARARGS     = enum.auto()
    MATCH_VARARGS    = enum.auto()

    MASK_IF_LIST     = enum.auto()
    MASK_IF_DICT     = enum.auto()
    
    FOCUS_ON_HEAD    = enum.auto()
    FOCUS_ON_KEY     = enum.auto()
    FOCUS_ON_BINDING = enum.auto()
    POP_FOCUS        = enum.auto()

    YIELD_BINDINGS   = enum.auto()

    def __repr__(self):
        return self.name